@@ -40,7 +40,7 @@ configure it, set the following options in your hgrc::
 """
 
 from mercurial.i18n import _
-from mercurial.node import *
+from mercurial.node import bin, short
 from mercurial import cmdutil, patch, templater, util, mail
 import email.Parser
 
@@ -113,7 +113,7 @@ class ciamsg(object):
 
         n = self.ctx.node()
         pbuf = patchbuf()
-        patch.export(self.cia.repo, [n], fp=pbuf)
+        cmdutil.export(self.cia.repo, [n], fp=pbuf)
         return patch.diffstat(pbuf.lines) or ''
 
     def logmsg(self):
@@ -197,7 +197,10 @@ class hgcia(object):
 
     def sendrpc(self, msg):
         srv = xmlrpclib.Server(self.ciaurl)
-        srv.hub.deliver(msg)
+        res = srv.hub.deliver(msg)
+        if res is not True and res != 'queued.':
+            raise util.Abort(_('%s returned an error: %s') %
+                             (self.ciaurl, res))
 
     def sendemail(self, address, data):
         p = email.Parser.Parser()
